#ifndef __mFrame_H__
#define __mFrame_H__

/*! 
    \file mFrame.h
    \brief 3D Axial Frame
    \author Douglas H. Cox
*/

//============================================================================
//    REQUIRED HEADERS
//============================================================================
#include "mMath.h"
#include "mVec3.h"

//============================================================================
//    DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS / CLASS FORWARDS
//============================================================================
class mQuat;

//============================================================================
//    FUNCTION PROTOTYPES
//============================================================================

//============================================================================
//    CLASS PROTOTYPES
//============================================================================

//! 3D Axial Frame
/*! mFrame represents a 3D Axial Frame.  An axial frame consists
    of three normalized, orthogonal vectors.  These describe a rotation
    value for a give coordinate system.
    
    Transformations \e from an axial frame perform a rotation
    from the frame's local-space into its parents.  Transformations
    \e to an axial frame perform a rotation from the frame's parent-
    space into its local-space.
    \sa http://www.loonygames.com/content/1.29/cotc/  */
class MMATH_DLL mFrame
{
public:
  mVec3 a[3]; //!< normalized x, y, z axes. right handed.

  //! Default constructor
  /*! No initialization is done here. */
  mFrame()
  {
  }

  //! Copy constructor
  mFrame(const mFrame& src)
  { 
    a[0].set(src.a[0]);
    a[1].set(src.a[1]);
    a[2].set(src.a[2]);
  }

  //! Constructor from 3 normalized, orthogonal axes.
  mFrame(const mVec3& x, const mVec3& y, const mVec3& z)
  {
    a[0].set(x);
    a[1].set(y);
    a[2].set(z);
  }

  //! Set this to cartesian axes.
  void makeIdentity()
  {
    a[0].set(1.0f, 0.0f, 0.0f);
    a[1].set(0.0f, 1.0f, 0.0f);
    a[2].set(0.0f, 0.0f, 1.0f);
  }

  //! Assignment
  void set(const mFrame& src)
  {
    a[0].set(src.a[0]);
    a[1].set(src.a[1]);
    a[2].set(src.a[2]);
  }

  //! Assignment from 3 normalized, orthogonal axes
  void set(const mVec3& x, const mVec3& y, const mVec3& z)
  {
    a[0].set(x);
    a[1].set(y);
    a[2].set(z);
  }

  //! Assignment
  mFrame& operator = (const mFrame& src)
  {
    a[0].set(src.a[0]);
    a[1].set(src.a[1]);
    a[2].set(src.a[2]);
    return *this;
  }

  //! Accessor
  void get(mVec3& x, mVec3& y, mVec3& z) const
  {
    x.set(a[0]);
    y.set(a[1]);
    z.set(a[2]);
  }

  //! Accessor
  mVec3& operator [] (int i)
  {
    return a[i];
  }

  //! Accessor
  const mVec3& operator [] (int i) const
  {
    return a[i];
  }

  //! Returns true if this == f
  bool operator == (const mFrame& f) const
  {
    return( a[0] == f.a[0] &&
            a[1] == f.a[1] &&
            a[2] == f.a[2] );
  }

  //! Returns true if this != f
  bool operator != (const mFrame& f) const
  {
    return !(operator==(f));
  }
  
  //! Returns true is this is within some tolerance value of \em f.
  bool approxEqual(const mFrame& f, float tol = mZeroTol) const;

  //! Computes the inverse of f. 
  void inverse();
  
  //! Computes the inverse of f. 
  void inverse(const mFrame& f);

  //! Transforms f into 'to's axial frame.
  void xformTo(const mFrame& f, const mFrame& to);

  //! Transforms f from 'from's axial frame.
  void xformFrom(const mFrame& f, const mFrame& from);
 
  //! Sets this to the rotation defined by the quaternion q.
  void setQuat(const mQuat& q);

  //! Gets a normalized quaternion.
  void getQuat(mQuat& q) const;

  //! Sets this to a rotation about a unit vector n. a is in degrees.
  void setAxisAngle(const mVec3& n, float a);

  //! Gets an angle about a unit vector n.
  void getAxisAngle(mVec3& n, float& a) const;

  //! Sets this to a rotation defined by the heading and pitch in degrees.
  /*! This assumes a roll of 0.0f and is faster than the full setHPR method. 
      \todo: Test this. */
  void setHP(float heading, float pitch);

  //! Sets this to a rotation defined by heading, pitch, and roll in degrees.
  /*! \todo: Test this. */
  void setHPR(float heading, float pitch, float roll);

  //! Sets this to a rotation defined by heading, pitch, and roll in degrees.
  void setHPR(const mVec3& hpr)
  {
    setHPR(hpr.v[0], hpr.v[1], hpr.v[2]);
  }

};

//****************************************************************************
//**    END HEADER mFrame.h
//****************************************************************************
#endif // __mFrame_H__
