#ifndef __mMath_H__
#define __mMath_H__

/*! 
    \file mMath.h 
    \brief Main mMath include file. 
    \author Douglas Cox 
*/

//=============================================================
//    DOCUMENTATION
//=============================================================

/*! 
  \mainpage mMath Library
  \section Introduction Introduction
  mMath is a 3d math library intend for games or real-time 3d Viz 
  applications.  It is still in development and is still being tested.  
  Currently it is based around a mFrame and mOCS class similar to 
  classes from Chris Hargrove's COTC articles.

  Optimizations have currently been kept at a portable level.  This
  may change in the future, but for now, please point out anything
  that you feel is non-optimal.  

  The vector classes are pretty complete.  If I have missed a commonly
  used function, please tell me.  The matrix class is meant to be
  very simple, because the OCS is meant to be used in most places
  that a matrix is normally used.

  The volume classes are still being worked on and will be added shortly.
  These currently include an axis-aligned bounding box, sphere, plane, 
  and frustum.  Ellipsoids, cylinders, and possibly an oriented bounding
  box will be added later.

  mMath will continue to be in the public domain.  All source will be
  provided each update.  If anyone wishes to help, please email me fixes
  or additions.

  <a href="http://www.doxygen.org>Doxygen</a> was used for all documentation.


  \section BeforeYouBegin Before You Begin
  Before you begin using any classes from the library, you should note
  the following:  mMath uses a right-handed, +Z = up, +Y = forward (into 
  the screen), +X = right coordinate system.  This is different than
  OpenGL (right-handed, +Y = up, -Z = forward, +X = right) and Direct3D 
  (left-handed, +Y = up, +Z = forward, +X = right).  Future examples
  will show how to convert to the desired API format.  
  
  Also, heading is defined as a rotation about the +Z axis, pitch is 
  a rotation about the +X axis, and roll is a rotation about the +Y axis.

  Finally, make sure you call mInitialize before calling any functions
  in this library.


  \section Compiling Compiling
  mMath.dsp compiles the library into a Win32 DLL.  If you wish to 
  compile the code into a large DLL or executable, #define MMATH_STATIC
  in your project's settings (Project / Settings / C++ / Preprocessor 
  definitions).
  This code should compile on non-Windows OSs.  If not, please email any
  corrections.


  \section Examples Examples
  Currently there is only a single, very simple example.  It briefly
  demonstrates how to use the mOCS class.  Other examples will be added
  shortly.  Please browse through the per-class documentation as well
  as the source code for further help.

*/
  

//============================================================================
//    REQUIRED HEADERS
//============================================================================

#ifdef WIN32
  #include <float.h>
#endif // WIN32

#include <assert.h>
#include <math.h>

//============================================================================
//    DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS / CLASS FORWARDS
//============================================================================

// Define MMATH_DLL for proper import/export behavior.
#ifndef WIN32
  #define MMATH_DLL
#else
  #ifdef MMATH_STATIC
  // ^ Define MMATH_STATIC to compile .cpp files statically 
  //   with another project.
    #define MMATH_DLL
  #else
    #ifdef MMATH_EXPORTS
    // ^ Should be defined by the project building the DLL. 
      #define MMATH_DLL __declspec(dllexport)
    #else
      #define MMATH_DLL __declspec(dllimport)
      #ifdef _DEBUG
        #pragma comment(lib, "mMath-dbg")
      #else
        #pragma comment(lib, "mMath")
      #endif
    #endif
  #endif
#endif

// Common constants

//! The smallest possible floating point value.
#define mFloatMin (FLT_MIN)

//! The largest possible floating point value.
#define mFloatMax (FLT_MAX)

//! PI
#define mPI 3.1415926536f

//! mPI / 2 
#define mHalfPI 1.5707963268f

//! A value < this is equal to zero.
#define mZeroTol 0.000001f

// Square root table used for fast square root function.
MMATH_DLL extern unsigned int mFastSqrtTable[0x10000];

// Macro used by fast sqrt function. (nVidia)
#define mFPBits(fp) (*(unsigned *)&(fp))

//============================================================================
//    FUNCTION PROTOTYPES
//============================================================================

//! Initializes the math library.
/*! This must be called before using any functions in this library. */
MMATH_DLL bool mInitialize();

//! Uninitializes the math library.
MMATH_DLL void mUninitialize();

//! Returns the absolute value of x.
inline float mAbs(float x) 
{
  return (x < 0.0f) ? -x : x;
}

//! Returns the smaller value.
template<class T>
inline T mMin(T a, T b)
{
  return (a < b ? a : b);
}

//! Returns the larger value.
template<class T>
inline T mMax(T a, T b)
{
  return (a > b ? a : b);
}

//! Returns the remainder of a / b.
inline float mMod(float a, float b)
{
  int n = int(a / b);
  a -= n * b;
  return (a < 0.0f) ? (a + b) : a;
}

//! Returns the floor of x.
inline float mFloor(float x)
{
  int ix = int(x);
  return float(ix - (x < 0.0f && x != float(ix)));
}

//! Returns the ceiling of x.
inline float mCeil(float x)
{
  int ix = int(x);
  return float(ix + (x > 0.0f && x != float(ix)));
}

//! Clamps x to a specific range [a, b].
inline float mClamp(float x, float a, float b)
{
  return (x < a ? a : (x > b ? b : x)); 
}

//! Returns true if a is approximately equal to b +- tol.
inline bool mApproxEqual(float a, float b, float tol = mZeroTol)
{
  return (mAbs(a-b) <= tol);
}

//! Rounds x to the closest integer.
inline int mRound(float x)
{
  return int(x + 0.5f);
}

//! Returns x * x.
inline float mSquare(float x)
{
  return (x*x);
}

//! Table-based square root of x. (nVidia)
inline float mSqrt(float x)
{
  if (mFPBits(x) == 0)
    return 0.0f;

  mFPBits(x) = mFastSqrtTable[(mFPBits(x) >> 8) & 0xFFFF] | 
    ((((mFPBits(x) - 0x3F800000) >> 1) + 0x3F800000) & 0x7F800000);
  
  return x;
}

//! Slower but more accurate version of mSqrt.
inline float mSlowSqrt(float x)
{
  return float(sqrt(x));
}


//! Converts radians to degrees
inline float mRadToDeg(float x) 
{
  return (x * 57.295779513f);
}

//! Converts degrees to radians
inline float mDegToRad(float x) 
{
  return (x * 0.017453292520f);
}

//! Calculates the cosine of a in degrees.
inline float mCos(float a)
{
  return float(cos(mDegToRad(a)));
}

//! Calculates the arc cosine of a in degrees.
inline float mACos(float a)
{
  return mRadToDeg(float(acos(mClamp(a, -1.0f, 1.0f))));
}

//! Calculates the sine of a in degrees.
inline float mSin(float a)
{
  return float(sin(mDegToRad(a)));
}

//! Calculates the arc sine of a in degrees.
inline float mASin(float a)
{
  return mRadToDeg(float(asin(mClamp(a, -1.0f, 1.0f))));
}

//! Calculates the sine and cosine of a in degrees.
inline void mSinCos(float a, float& sine, float& cosine)
{
  a = mDegToRad(a);
  sine = float(sin(a));
  cosine = float(cos(a));

  // Don't Intel CPUs have a SINCOS op?
}

//! Calculates the tangent of a in degrees.
inline float mTan(float a)
{
  return float(tan(mDegToRad(a))); 
}

//! Calculates the arc tangent of x in degrees.
inline float mATan(float x)
{
  return mRadToDeg(float(atan(x))); 
}

//! Calculates the arc tangent of y/x in degrees.
inline float mATan2(float y, float x)
{
  return mRadToDeg(float(atan2(y, x)));
}

//============================================================================
//    CLASS PROTOTYPES
//============================================================================

//****************************************************************************
//**    END HEADER mMath.h
//****************************************************************************
#endif // __mMath_H__
